#include "FirstOrderTerm.h"
#include "DOFVector.h"

namespace AMDiS {

  // =========== VecAtQP_FOT ==========

  VecAtQP_FOT::VecAtQP_FOT(DOFVectorBase<double> *dv,
			   AbstractFunction<double, double> *af,
			   WorldVector<double> *wv)
    : FirstOrderTerm(af->getDegree()), vec(dv), f(af), b(wv)
  {
    TEST_EXIT(dv)("No vector!\n");

    auxFeSpaces.insert(dv->getFESpace());
  }

  VecAtQP_FOT::VecAtQP_FOT(DOFVectorBase<double> *dv,
			   AbstractFunction<double, double> *af,
			   int bIdx)
    : FirstOrderTerm(af->getDegree()), vec(dv), f(af)
  {
    TEST_EXIT(dv)("No vector!\n");

    bOne = bIdx;
    auxFeSpaces.insert(dv->getFESpace());
  }

  void VecAtQP_FOT::initElement(const ElInfo* elInfo, 
				SubAssembler* subAssembler,
				Quadrature *quad) 
  {
    vecAtQPs = getVectorAtQPs(vec, elInfo, subAssembler, quad);
  }

  void VecAtQP_FOT::initElement(const ElInfo* smallElInfo,
				const ElInfo* largeElInfo,
				SubAssembler* subAssembler,
				Quadrature *quad)
  {
    vecAtQPs = getVectorAtQPs(vec, smallElInfo, largeElInfo, subAssembler, quad);
  }

  void VecAtQP_FOT::getLb(const ElInfo *elInfo, int nPoints, 
			  VectorOfFixVecs<DimVec<double> >& Lb) const 
  {
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();

    if (bOne > -1) {
      for (int iq = 0; iq < nPoints; iq++)
	lb_one(Lambda, Lb[iq], (*f)(vecAtQPs[iq]));
    } else if (b) {
      for (int iq = 0; iq < nPoints; iq++)
	lb(Lambda, *b, Lb[iq], (*f)(vecAtQPs[iq]));
    } else {
      for (int iq = 0; iq < nPoints; iq++)
	l1(Lambda, Lb[iq], (*f)(vecAtQPs[iq]));      
    }
  }

  void VecAtQP_FOT::eval(int nPoints,
			 const double *uhAtQP,
			 const WorldVector<double> *grdUhAtQP,
			 const WorldMatrix<double> *D2UhAtQP,
			 double *result,
			 double fac) 
  {
    int dow = Global::getGeo(WORLD);
  
    if (grdUhAtQP) {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = (*f)(vecAtQPs[iq]);
	double resultQP = 0.0;
	for (int i = 0; i < dow; i++)
	  resultQP += grdUhAtQP[iq][i];
		
	result[iq] += fac * factor * resultQP;
      } 
    }
  }

  
  // =========== CoordsAtQP_FOT ===========

  void CoordsAtQP_FOT::initElement(const ElInfo* elInfo, 
				   SubAssembler* subAssembler,
				   Quadrature *quad) 
  {
    coordsAtQPs = subAssembler->getCoordsAtQPs(elInfo, quad);
  }

  void CoordsAtQP_FOT::getLb(const ElInfo *elInfo, int nPoints, 
			     VectorOfFixVecs<DimVec<double> >& Lb) const 
  {
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    for (int iq = 0; iq < nPoints; iq++)
      l1(Lambda, Lb[iq], (*g)(coordsAtQPs[iq]));
  }

  void CoordsAtQP_FOT::eval(int nPoints,
			    const double *uhAtQP,
			    const WorldVector<double> *grdUhAtQP,
			    const WorldMatrix<double> *D2UhAtQP,
			    double *result,
			    double f) 
  {
    int dow = Global::getGeo(WORLD);
  
    if (grdUhAtQP) {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = (*g)(coordsAtQPs[iq]);
	double resultQP = 0.0;
	for (int i = 0; i < dow; i++)
	  resultQP += grdUhAtQP[iq][i];	
	result[iq] += f * factor * resultQP;
      }
    }
  }


  // ========== VecCoordsAtQP_FOT ==========

  void VecCoordsAtQP_FOT::initElement(const ElInfo* elInfo, 
				      SubAssembler* subAssembler,
				      Quadrature *quad) 
  {
    coordsAtQPs = subAssembler->getCoordsAtQPs(elInfo, quad);
  }

  void VecCoordsAtQP_FOT::getLb(const ElInfo *elInfo, int nPoints,
				VectorOfFixVecs<DimVec<double> >& Lb) const
  {
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    for (int iq = 0; iq < nPoints; iq++)
      lb(Lambda, b, Lb[iq], (*g)(coordsAtQPs[iq]));
  }

  void VecCoordsAtQP_FOT::eval(int nPoints,
			       const double *uhAtQP,
			       const WorldVector<double> *grdUhAtQP,
			       const WorldMatrix<double> *D2UhAtQP,
			       double *result,
			       double f) 
  {
    int dow = Global::getGeo(WORLD);
  
    if (grdUhAtQP) {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = (*g)(coordsAtQPs[iq]);
	double resultQP = 0.0;
	for (int i = 0; i < dow; i++)
	  resultQP += grdUhAtQP[iq][i];
	
	result[iq] += f * factor * resultQP;
      }
    }
  }


  // ========== VectorGradient_FOT ==========

  VectorGradient_FOT::VectorGradient_FOT(DOFVectorBase<double> *dv,
					 AbstractFunction<WorldVector<double>, WorldVector<double> > *af)
    : FirstOrderTerm(af->getDegree()), vec(dv), f(af)
  {
    TEST_EXIT(dv)("No vector!\n");

    auxFeSpaces.insert(dv->getFESpace());
  }

  void VectorGradient_FOT::initElement(const ElInfo* elInfo, 
				       SubAssembler* subAssembler,
				       Quadrature *quad) 
  {
    gradAtQPs = getGradientsAtQPs(vec, elInfo, subAssembler, quad);
  }

  void VectorGradient_FOT::getLb(const ElInfo *elInfo, int nPoints, 
				 VectorOfFixVecs<DimVec<double> >& Lb) const 
  {
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    if (f) {
      for (int iq = 0; iq < nPoints; iq++)
	lb(Lambda, (*f)(gradAtQPs[iq]), Lb[iq], 1.0);      
    } else {
      for (int iq = 0; iq < nPoints; iq++)
	lb(Lambda, gradAtQPs[iq], Lb[iq], 1.0);      
    }
  }

  void VectorGradient_FOT::eval(int nPoints,
				const double *uhAtQP,
				const WorldVector<double> *grdUhAtQP,
				const WorldMatrix<double> *D2UhAtQP,
				double *result,
				double factor) 
  { 
    if (grdUhAtQP) {
      if (f) {
	for (int iq = 0; iq < nPoints; iq++) {
	  WorldVector<double> b = (*f)(gradAtQPs[iq]);
	  result[iq] += b * grdUhAtQP[iq] * factor;
	}
      } else {
	for (int iq = 0; iq < nPoints; iq++)
	  result[iq] += gradAtQPs[iq] * grdUhAtQP[iq] * factor;	
      }
    }
  }


  // =========== VectorFct_FOT ==========

  VectorFct_FOT::VectorFct_FOT(DOFVectorBase<double> *dv,
			       AbstractFunction<WorldVector<double>, double> *fct)
    : FirstOrderTerm(fct->getDegree()), vec(dv), vecFct(fct) 
  {
    TEST_EXIT(dv)("No vector!\n");

    auxFeSpaces.insert(dv->getFESpace());
  }

  void VectorFct_FOT::initElement(const ElInfo* elInfo, 
				  SubAssembler* subAssembler,
				  Quadrature *quad) 
  {
    vecAtQPs = getVectorAtQPs(vec, elInfo, subAssembler, quad);
  }

  void VectorFct_FOT::getLb(const ElInfo *elInfo, int nPoints, 
			    VectorOfFixVecs<DimVec<double> >& Lb) const 
  {
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    for (int iq = 0; iq < nPoints; iq++)
      lb(Lambda, (*vecFct)(vecAtQPs[iq]), Lb[iq], 1.0);    
  }

  void VectorFct_FOT::eval(int nPoints,
			   const double *uhAtQP,
			   const WorldVector<double> *grdUhAtQP,
			   const WorldMatrix<double> *D2UhAtQP,
			   double *result,
			   double factor) 
  {
    if (grdUhAtQP) {
      for (int iq = 0; iq < nPoints; iq++) {
	WorldVector<double> b = (*vecFct)(vecAtQPs[iq]);
	result[iq] += b * grdUhAtQP[iq] * factor;
      }
    }
  }


  // =========== VecFctAtQP_FOT ==========
  
  void VecFctAtQP_FOT::initElement(const ElInfo* elInfo, 
				   SubAssembler* subAssembler,
				   Quadrature *quad) 
  {
    coordsAtQPs = subAssembler->getCoordsAtQPs(elInfo, quad);
  }


  void VecFctAtQP_FOT::getLb(const ElInfo *elInfo, int nPoints, 
			     VectorOfFixVecs<DimVec<double> >& Lb) const 
  {
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    for (int iq = 0; iq < nPoints; iq++)
      lb(Lambda, (*g)(coordsAtQPs[iq]), Lb[iq], 1.0);    
  }

  void VecFctAtQP_FOT::eval(int nPoints,
			    const double *uhAtQP,
			    const WorldVector<double> *grdUhAtQP,
			    const WorldMatrix<double> *D2UhAtQP,
			    double *result,
			    double f) 
  {
    int dow = Global::getGeo(WORLD);
  
    if (grdUhAtQP) {
      for (int iq = 0; iq < nPoints; iq++) {
	double resultQP = 0.0;
	const WorldVector<double> &b = (*g)(coordsAtQPs[iq]);
	for (int i = 0; i < dow; i++)
	  resultQP += b[i] * grdUhAtQP[iq][i];	
	result[iq] += f * resultQP;
      }
    }
  }


  // =========== VecGrad_FOT ===========

  VecGrad_FOT::VecGrad_FOT(DOFVectorBase<double> *dv1, DOFVectorBase<double> *dv2,
			   BinaryAbstractFunction<WorldVector<double>, double, WorldVector<double> > *fct)
    : FirstOrderTerm(fct->getDegree()), vec1(dv1), vec2(dv2), vecFct(fct) 
  {
    TEST_EXIT(dv1)("No first vector!\n");
    TEST_EXIT(dv2)("No second vector!\n");

    auxFeSpaces.insert(dv1->getFESpace());
    auxFeSpaces.insert(dv2->getFESpace());    
  }

  void VecGrad_FOT::initElement(const ElInfo* elInfo, 
				SubAssembler* subAssembler,
				Quadrature *quad) 
  {
    vecAtQPs = getVectorAtQPs(vec1, elInfo, subAssembler, quad);
    gradAtQPs = getGradientsAtQPs(vec2, elInfo, subAssembler, quad);
  }

  void VecGrad_FOT::initElement(const ElInfo* smallElInfo,
				const ElInfo* largeElInfo,
				SubAssembler* subAssembler,
				Quadrature *quad)
  {
    vecAtQPs = getVectorAtQPs(vec1, smallElInfo, largeElInfo, subAssembler, quad);
    gradAtQPs = getGradientsAtQPs(vec2, smallElInfo, largeElInfo, subAssembler, quad);
  }

  void VecGrad_FOT::getLb(const ElInfo *elInfo, int nPoints, 
			  VectorOfFixVecs<DimVec<double> >& Lb) const {
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    for (int iq = 0; iq < nPoints; iq++)
      lb(Lambda, (*vecFct)(vecAtQPs[iq], gradAtQPs[iq]), Lb[iq], 1.0);
  }
  
  void VecGrad_FOT::eval(int nPoints,
			 const double *uhAtQP,
			 const WorldVector<double> *grdUhAtQP,
			 const WorldMatrix<double> *D2UhAtQP,
			 double *result,
			 double factor) 
  {
    if (grdUhAtQP) {
      for (int iq = 0; iq < nPoints; iq++) {
	WorldVector<double> b = (*vecFct)(vecAtQPs[iq], gradAtQPs[iq]);
	result[iq] += b * grdUhAtQP[iq] * factor;
      }
    }
  }


  // =========== FctGrad2_FOT ===========

  void FctGrad2_FOT::initElement(const ElInfo* elInfo, 
				SubAssembler* subAssembler,
				Quadrature *quad) 
  {
    grad1AtQPs = subAssembler->getGradientsAtQPs(vec1, elInfo, quad);
    grad2AtQPs = subAssembler->getGradientsAtQPs(vec2, elInfo, quad);
  }

  void FctGrad2_FOT::getLb(const ElInfo *elInfo, int nPoints, 
			   VectorOfFixVecs<DimVec<double> >& Lb) const 
  {
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    for (int iq = 0; iq < nPoints; iq++)
      lb(Lambda, (*vecFct)(grad1AtQPs[iq], grad2AtQPs[iq]), Lb[iq], 1.0);
  }
  
  void FctGrad2_FOT::eval(int nPoints,
			  const double *uhAtQP,
			  const WorldVector<double> *grdUhAtQP,
			  const WorldMatrix<double> *D2UhAtQP,
			  double *result,
			  double factor) 
  {
    if (grdUhAtQP) {
      for (int iq = 0; iq < nPoints; iq++) {
	WorldVector<double> b = (*vecFct)(grad1AtQPs[iq], grad2AtQPs[iq]);
	result[iq] += b * grdUhAtQP[iq] * factor;
      }
    }
  }


  // ========== Vec2AndGradAtQP_FOT ==========

  Vec2AndGradAtQP_FOT::Vec2AndGradAtQP_FOT(DOFVectorBase<double> *dv1, 
					   DOFVectorBase<double> *dv2,
					   TertiaryAbstractFunction<double, double, double, WorldVector<double> > *f_, WorldVector<double> *b_)
    : FirstOrderTerm(f_->getDegree()), 
      vec1(dv1), 
      vec2(dv2), 
      f(f_), 
      b(b_)
  {   
    auxFeSpaces.insert(dv1->getFESpace()); 
    auxFeSpaces.insert(dv2->getFESpace());
  }

  void Vec2AndGradAtQP_FOT::initElement(const ElInfo* elInfo, 
					SubAssembler* subAssembler,
					Quadrature *quad) 
  { 
    vec1AtQPs = subAssembler->getVectorAtQPs(vec1, elInfo, quad);
    vec2AtQPs = subAssembler->getVectorAtQPs(vec2, elInfo, quad);
    gradAtQPs1 = getGradientsAtQPs(vec1, elInfo, subAssembler, quad);
  }
    
  void Vec2AndGradAtQP_FOT::getLb(const ElInfo *elInfo, int nPoints, VectorOfFixVecs<DimVec<double> >& Lb) const {
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    for (int iq = 0; iq < nPoints; iq++) {
      if (b)
	lb(Lambda, *b, Lb[iq], (*f)(vec1AtQPs[iq], vec2AtQPs[iq],gradAtQPs1[iq]));
      else
	l1(Lambda, Lb[iq], (*f)(vec1AtQPs[iq], vec2AtQPs[iq], gradAtQPs1[iq]));
    }
  }
  
  void Vec2AndGradAtQP_FOT::eval(int nPoints,
				 const double *uhAtQP,
				 const WorldVector<double> *grdUhAtQP,
				 const WorldMatrix<double> *D2UhAtQP,
				 double *result,
				 double fac) 
  {   
    if (grdUhAtQP)
      for (int iq = 0; iq < nPoints; iq++)
	result[iq] += fac * 
	  (*f)(vec1AtQPs[iq], vec2AtQPs[iq], gradAtQPs1[iq]) * 
	  ((*b) * grdUhAtQP[iq]);
  }


  // ========== FctVecAtQP_FOT ==========


  FctVecAtQP_FOT::FctVecAtQP_FOT(DOFVectorBase<double> *dv,
				 BinaryAbstractFunction<double, WorldVector<double>, double> *f_,
				 WorldVector<double> *b_)
    : FirstOrderTerm(f_->getDegree()), vec(dv), f(f_), b(b_)
  {
    auxFeSpaces.insert(dv->getFESpace());
  }

  void FctVecAtQP_FOT::initElement(const ElInfo* elInfo, 
				   SubAssembler* subAssembler,
				   Quadrature *quad) 
  { 
    vecAtQPs = subAssembler->getVectorAtQPs(vec, elInfo, quad);
    coordsAtQPs = subAssembler->getCoordsAtQPs(elInfo, quad);
  }

  void FctVecAtQP_FOT::getLb(const ElInfo *elInfo, int nPoints, 
			     VectorOfFixVecs<DimVec<double> >& Lb) const 
  {
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    for (int iq = 0; iq < nPoints; iq++) {
      if (b)
	lb(Lambda, *b, Lb[iq], (*f)(coordsAtQPs[iq], vecAtQPs[iq]));
      else
	l1(Lambda, Lb[iq], (*f)(coordsAtQPs[iq], vecAtQPs[iq]));
    }
  }
  
  void FctVecAtQP_FOT::eval(int nPoints,
			    const double *uhAtQP,
			    const WorldVector<double> *grdUhAtQP,
			    const WorldMatrix<double> *D2UhAtQP,
			    double *result,
			    double fac) 
  {
    if (grdUhAtQP)
      for (int iq = 0; iq < nPoints; iq++)
	result[iq] += 
	  fac * (*f)(coordsAtQPs[iq], vecAtQPs[iq]) * ((*b) * grdUhAtQP[iq]);
  }


  // ========== Vec2AtQP_FOT ==========

  Vec2AtQP_FOT::Vec2AtQP_FOT(DOFVectorBase<double> *dv1, DOFVectorBase<double> *dv2,
			     BinaryAbstractFunction<double, double, double> *af,
			     WorldVector<double> *b_)
    : FirstOrderTerm(af->getDegree()), vec1(dv1), vec2(dv2), f(af), b(b_)
  {   
    TEST_EXIT(dv1)("No first vector!\n");
    TEST_EXIT(dv2)("No second vector!\n");

    auxFeSpaces.insert(dv1->getFESpace()); 
    auxFeSpaces.insert(dv2->getFESpace());
  }

  Vec2AtQP_FOT::Vec2AtQP_FOT(DOFVectorBase<double> *dv1, DOFVectorBase<double> *dv2,
			     BinaryAbstractFunction<double, double, double> *af,
			     int bIdx)
    : FirstOrderTerm(af->getDegree()), vec1(dv1), vec2(dv2), f(af)
  {   
    TEST_EXIT(dv1)("No first vector!\n");
    TEST_EXIT(dv2)("No second vector!\n");

    bOne = bIdx;
    auxFeSpaces.insert(dv1->getFESpace()); 
    auxFeSpaces.insert(dv2->getFESpace());
  }

  void Vec2AtQP_FOT::initElement(const ElInfo* elInfo, 
				 SubAssembler* subAssembler,
				 Quadrature *quad) 
  { 
    vec1AtQPs = subAssembler->getVectorAtQPs(vec1, elInfo, quad);
    vec2AtQPs = subAssembler->getVectorAtQPs(vec2, elInfo, quad);
  }
    
  void Vec2AtQP_FOT::getLb(const ElInfo *elInfo, int nPoints, 
			   VectorOfFixVecs<DimVec<double> >& Lb) const 
  {
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();

    if (bOne > -1) {
      for (int iq = 0; iq < nPoints; iq++)
	lb_one(Lambda, Lb[iq], (*f)(vec1AtQPs[iq], vec2AtQPs[iq]));
    } else if (b) {
      for (int iq = 0; iq < nPoints; iq++)
	lb(Lambda, *b, Lb[iq], (*f)(vec1AtQPs[iq], vec2AtQPs[iq]));
    } else {
      for (int iq = 0; iq < nPoints; iq++)
	l1(Lambda, Lb[iq], (*f)(vec1AtQPs[iq], vec2AtQPs[iq]));
    }
  }
  
  void Vec2AtQP_FOT::eval(int nPoints,
			  const double *uhAtQP,
			  const WorldVector<double> *grdUhAtQP,
			  const WorldMatrix<double> *D2UhAtQP,
			  double *result,
			  double fac) 
  {
    if (grdUhAtQP)
      for (int iq = 0; iq < nPoints; iq++) 
	result[iq] += fac * (*f)(vec1AtQPs[iq], vec2AtQPs[iq]) * ((*b) * grdUhAtQP[iq]);
  }

  
  // ========== Vec3FctAtQP_FOT ==========

  Vec3FctAtQP_FOT::Vec3FctAtQP_FOT(DOFVectorBase<double> *dv1, 
				   DOFVectorBase<double> *dv2, 
				   DOFVectorBase<double> *dv3,
				   TertiaryAbstractFunction<double, double, double, double> *f_,
				   WorldVector<double> *bvec)
    : FirstOrderTerm(f_->getDegree()), 
      vec1(dv1), 
      vec2(dv2), 
      vec3(dv3), 
      f(f_),
      b(bvec)
  { 
    auxFeSpaces.insert(dv1->getFESpace()); 
    auxFeSpaces.insert(dv2->getFESpace()); 
    auxFeSpaces.insert(dv3->getFESpace());   
  }

  Vec3FctAtQP_FOT::Vec3FctAtQP_FOT(DOFVectorBase<double> *dv1, 
				   DOFVectorBase<double> *dv2, 
				   DOFVectorBase<double> *dv3,
				   TertiaryAbstractFunction<double, double, double, double> *f_,
				   int b)
    : FirstOrderTerm(f_->getDegree()), 
      vec1(dv1), 
      vec2(dv2), 
      vec3(dv3), 
      f(f_)
  { 
    bOne = b;
    auxFeSpaces.insert(dv1->getFESpace()); 
    auxFeSpaces.insert(dv2->getFESpace()); 
    auxFeSpaces.insert(dv3->getFESpace());   
  }

  void Vec3FctAtQP_FOT::initElement(const ElInfo* elInfo, 
				    SubAssembler* subAssembler,
				    Quadrature *quad) 
  { 
    vec1AtQPs = subAssembler->getVectorAtQPs(vec1, elInfo, quad);
    vec2AtQPs = subAssembler->getVectorAtQPs(vec2, elInfo, quad);
    vec3AtQPs = subAssembler->getVectorAtQPs(vec3, elInfo, quad);
  }
  
  void Vec3FctAtQP_FOT::getLb(const ElInfo *elInfo, int nPoints, 
			      VectorOfFixVecs<DimVec<double> >& Lb) const 
  {
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    if (bOne > -1) {
      for (int iq = 0; iq < nPoints; iq++)
	lb_one(Lambda, Lb[iq], (*f)(vec1AtQPs[iq], vec2AtQPs[iq], vec3AtQPs[iq]));
    } else {
      for (int iq = 0; iq < nPoints; iq++) {
	if (b)
	  lb(Lambda, *b, Lb[iq], (*f)(vec1AtQPs[iq], vec2AtQPs[iq], vec3AtQPs[iq]));
	else
	  l1(Lambda, Lb[iq], (*f)(vec1AtQPs[iq], vec2AtQPs[iq], vec3AtQPs[iq]));
      }
    }
  }

  void Vec3FctAtQP_FOT::eval(int nPoints,
			     const double *uhAtQP,
			     const WorldVector<double> *grdUhAtQP,
			     const WorldMatrix<double> *D2UhAtQP,
			     double *result,
			     double fac) 
  {
    if (grdUhAtQP)
      for (int iq = 0; iq < nPoints; iq++)
	result[iq] += fac * (*f)(vec1AtQPs[iq], vec2AtQPs[iq] ,vec3AtQPs[iq]) * 
	  ((*b) * grdUhAtQP[iq]);
  }


  // =========== General_FOT ==========

  General_FOT::General_FOT(std::vector<DOFVectorBase<double>*> vecs,
			   std::vector<DOFVectorBase<double>*> grads,
			   TertiaryAbstractFunction<WorldVector<double>, 
			   WorldVector<double>,
			   std::vector<double>, 
			   std::vector<WorldVector<double> > > *af)
    : FirstOrderTerm(af->getDegree()), vecs_(vecs), grads_(grads), f_(af)
  {
    vecsAtQPs_.resize(vecs_.size());
    gradsAtQPs_.resize(grads_.size());

    for (int i = 0; i < static_cast<int>(vecs.size()); i++) {
      TEST_EXIT(vecs[i])("One vector is NULL!\n");

      auxFeSpaces.insert(vecs[i]->getFESpace());
    }   

    for (int i = 0; i < static_cast<int>(grads.size()); i++) {
      TEST_EXIT(grads[i])("One gradient vector is NULL!\n");

      auxFeSpaces.insert(grads[i]->getFESpace());
    }   
  }

  void General_FOT::initElement(const ElInfo* elInfo, 
				SubAssembler* subAssembler,
				Quadrature *quad)
  {
    int nVecs = static_cast<int>(vecs_.size());
    int nGrads = static_cast<int>(grads_.size());

    coordsAtQPs_ = subAssembler->getCoordsAtQPs(elInfo, quad);

    for (int i = 0; i < nVecs; i++)
      vecsAtQPs_[i] = getVectorAtQPs(vecs_[i], elInfo, subAssembler, quad);    
    for (int i = 0; i < nGrads; i++)
      gradsAtQPs_[i] = getGradientsAtQPs(grads_[i], elInfo, subAssembler, quad);    
  }

  void General_FOT::getLb(const ElInfo *elInfo, 
			  int nPoints, 
			  VectorOfFixVecs<DimVec<double> >& result) const
  {
    int nVecs = static_cast<int>(vecs_.size());
    int nGrads = static_cast<int>(grads_.size());

    std::vector<double> vecsArg(nVecs);
    std::vector<WorldVector<double> > gradsArg(nGrads);

    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();

    for (int iq = 0; iq < nPoints; iq++) {
      for (int i = 0; i < nVecs; i++)
	vecsArg[i] = vecsAtQPs_[i][iq];      
      for (int i = 0; i < nGrads; i++) 
	gradsArg[i] = gradsAtQPs_[i][iq];      

      lb(Lambda, (*f_)(coordsAtQPs_[iq], vecsArg, gradsArg), 
	 result[iq], 1.0);
    }
  }

  void General_FOT::eval(int nPoints,
			 const double *uhAtQP,
			 const WorldVector<double> *grdUhAtQP,
			 const WorldMatrix<double> *D2UhAtQP,
			 double *result,
			 double factor) 
  {
    int dow = Global::getGeo(WORLD);  
    int nVecs = static_cast<int>(vecs_.size());
    int nGrads = static_cast<int>(grads_.size());

    std::vector<double> vecsArg(nVecs);
    std::vector<WorldVector<double> > gradsArg(nGrads);

    if (grdUhAtQP) {
      for (int iq = 0; iq < nPoints; iq++) {
	double resultQP = 0.0;

	for (int i = 0; i < nVecs; i++)
	  vecsArg[i] = vecsAtQPs_[i][iq];	
	for (int i = 0; i < nGrads; i++)
	  gradsArg[i] = gradsAtQPs_[i][iq];	

	const WorldVector<double> &b = (*f_)(coordsAtQPs_[iq], vecsArg, gradsArg);

	for (int i = 0; i < dow; i++) 
	  resultQP += grdUhAtQP[iq][i] * b[i];
	
	result[iq] += factor * resultQP;
      }
    }
  }


  // =========== GeneralParametric_FOT ==========

  GeneralParametric_FOT::GeneralParametric_FOT(std::vector<DOFVectorBase<double>*> vecs,
					       std::vector<DOFVectorBase<double>*> grads,
					       QuartAbstractFunction<WorldVector<double>, 
					       WorldVector<double>,
					       WorldVector<double>,
					       std::vector<double>, 
					       std::vector<WorldVector<double> > > *af)
    : FirstOrderTerm(af->getDegree()), vecs_(vecs), grads_(grads), f_(af)
  {
    vecsAtQPs_.resize(vecs_.size());
    gradsAtQPs_.resize(grads_.size());

    for (int i = 0; i < static_cast<int>(vecs.size()); i++) {
      TEST_EXIT(vecs[i])("One vector is NULL!\n");

      auxFeSpaces.insert(vecs[i]->getFESpace());
    }   

    for (int i = 0; i < static_cast<int>(grads.size()); i++) {
      TEST_EXIT(grads[i])("One gradient vector is NULL!\n");

      auxFeSpaces.insert(grads[i]->getFESpace());
    }   
  }


  void GeneralParametric_FOT::initElement(const ElInfo* elInfo, 
					  SubAssembler* subAssembler,
					  Quadrature *quad)
  {
    int nVecs = static_cast<int>(vecs_.size());
    int nGrads = static_cast<int>(grads_.size());

    elInfo->getElementNormal(elementNormal_);
    coordsAtQPs_ = subAssembler->getCoordsAtQPs(elInfo, quad);

    for (int i = 0; i < nVecs; i++)
      vecsAtQPs_[i] = getVectorAtQPs(vecs_[i], elInfo, subAssembler, quad);    
    for (int i = 0; i < nGrads; i++) 
      gradsAtQPs_[i] = getGradientsAtQPs(grads_[i], elInfo, subAssembler, quad);    
  }

  void GeneralParametric_FOT::getLb(const ElInfo *elInfo, 
				    int nPoints, 
				    VectorOfFixVecs<DimVec<double> >& result) const
  {
    int nVecs = static_cast<int>(vecs_.size());
    int nGrads = static_cast<int>(grads_.size());

    std::vector<double> vecsArg(nVecs);
    std::vector<WorldVector<double> > gradsArg(nGrads);

    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    for (int iq = 0; iq < nPoints; iq++) {
      for (int i = 0; i < nVecs; i++)
	vecsArg[i] = vecsAtQPs_[i][iq];      
      for (int i = 0; i < nGrads; i++)
	gradsArg[i] = gradsAtQPs_[i][iq];
      
      lb(Lambda, (*f_)(coordsAtQPs_[iq],  elementNormal_, vecsArg, gradsArg), result[iq], 1.0);
    }
  }

  void GeneralParametric_FOT::eval(int nPoints,
				   const double *uhAtQP,
				   const WorldVector<double> *grdUhAtQP,
				   const WorldMatrix<double> *D2UhAtQP,
				   double *result,
				   double factor) 
  {
    int dow = Global::getGeo(WORLD);
    int nVecs = static_cast<int>(vecs_.size());
    int nGrads = static_cast<int>(grads_.size());

    std::vector<double> vecsArg(nVecs);
    std::vector<WorldVector<double> > gradsArg(nGrads);

    if (grdUhAtQP) {
      for (int iq = 0; iq < nPoints; iq++) {
	double resultQP = 0.0;

	for (int i = 0; i < nVecs; i++)
	  vecsArg[i] = vecsAtQPs_[i][iq];
	for (int i = 0; i < nGrads; i++)
	  gradsArg[i] = gradsAtQPs_[i][iq];	

	const WorldVector<double> &b = (*f_)(coordsAtQPs_[iq],  elementNormal_, vecsArg, gradsArg);

	for (int i = 0; i < dow; i++)
	  resultQP += grdUhAtQP[iq][i] * b[i];
	
	result[iq] += factor * resultQP;
      }
    }
  }



}
